import numpy as np
from tensorflow import keras as ks


def setup():

    print("Core setup done!")


class Core(object):

    """Core module (CosmoHammer)"""

    def __init__(self, load_model, norm=1.0):
        """
        :param load_model: load the model(can be a function of parameters model=f(a,b,c)),
        :param norm: the rescaling used in training, if any, default = 1.0.
        """
        self.model = ks.models.load_model(load_model)
        self.norm = norm

    def __call__(self, ctx):
        """
        :param ctx: contexts saves proposed step in (parameter, data) space
        """

        params = ctx.getParams()

        ctx.add('params_pk', params)

        mhmin, nion, rmfp = params

        params = np.array([[mhmin, nion, rmfp]])

        params = np.reshape(params, (1, 3))

        model_th = self.model.predict(params)
        model_th = model_th * self.norm

        ctx.add("model_th", model_th)

    @staticmethod
    def setup():
        print('core is done!')


import os
import time
from EmuPBk.MCMC.core import Core
from EmuPBk.MCMC.like import LikeModule
from cosmoHammer.util import Params
from cosmoHammer import MpiCosmoHammerSampler
from cosmoHammer import CosmoHammerSampler
from cosmoHammer import LikelihoodComputationChain
#  from cosmoHammer.pso.MpiParticleSwarmOptimizer import MpiParticleSwarmOptimizer

# The parameter space is defined
# paramters = [peak, min., max., jump] ===> A rough idea about the prior

path = os.path.abspath(os.path.join(__file__, os.pardir))
path = path + '/existing_models/'
params = Params(("NoH", [275, 10, 550, 3]),
                ("n_ion", [90.00, 10.00, 180.00, 1]),
                ("R_mfp", [30.00, 5.00, 60.00, 0.5]))


class RunMCMC:

    """ sampler & MPI sampler class """
    def __init__(self, data, nbins, noise=0., div=1.0, name='pk'):
        """
        :param data: load your data
        :param nbins: number of k-modes in powerspectrum OR
         number of triangle contributions in bispectrum (for covariance matrix)
        :param noise: system noise, e.g. SKA, MWA noise response (if any), default 0.0,
        :param div: likelihood normalization factor, default 1.0,
        :param name: use ('pk','bk') for powerspectrum, bispectrum
        """
        chain = LikelihoodComputationChain(min=params[:, 1], max=params[:, 2])
        chain.params = params
        chain.addLikelihoodModule(LikeModule(data, nbins, noise, div))
        self.chain = chain
        self.name = name

    def load_existing_model(self):
        """
        Use the existing ANN models for MCMC analysis
        """

        if self.name == 'pk':
            self.chain.addCoreModule(Core(load_model=path+'pk.h5', norm=1.))

        elif self.name == 'bk':
            self.chain.addCoreModule(Core(load_model=path+'bk.h5', norm=100.))

        self.chain.setup()

    def load_model(self, load_model, name='pk', norm=1.):

        """
        :param load_model: load your own model, (give the path)
        :param name: name for data, ('pk','bk')==>for powerspectrum, bispectrum
        :param norm: rescale used in the training
        """
        self.name = name
        self.chain.addCoreModule(Core(load_model, norm))
        self.chain.setup()

    def mpi_sampler(self, walker_ratio, burnin, samples, num, threads=-1):
        """
        :param walker_ratio:  the ratio of walkers and the count of sampled parameters
        :param burnin: burin iterations
        :param samples: no. of sample iterations
        :param num: number to put in output files e.g: string(name+num)=Pk_1,Bk_1
        :param threads: no. of cpu threads
        """
        #   self.chain.setup()
        #   print("find best fit point")
        #   pso = MpiParticleSwarmOptimizer(self.chain, params[:, 1], params[:, 2])
        #   psoTrace = np.array([pso.gbest.position.copy() for _ in pso.sample()])
        #   params[:, 0] = pso.gbest.position
        mpi_sampler = MpiCosmoHammerSampler(params=params,
                                            likelihoodComputationChain=self.chain,
                                            filePrefix='%s' % self.name+'%d' % num,
                                            walkersRatio=walker_ratio,
                                            burninIterations=burnin,
                                            sampleIterations=samples, threadCount=threads)
        print("started sampling:")
        start = time.time()
        mpi_sampler.startSampling()
        end = time.time()
        tics = end - start
        print("The time taken %.2f sec. done!" % tics)
        print('Done!')

    def sampler(self, walker_ratio, burnin, samples, num, threads=-1):
        """
            :param walker_ratio:  the ratio of walkers and the count of sampled parameters
            :param burnin: burin iterations
            :param samples: no. of sample iterations
            :param num: number to put in output files e.g: string(name+num)=Pk_1,Bk_1
            :param threads: no. of cpu threads

            self.chain.setup()
            print("find best fit point")
            pso = MpiParticleSwarmOptimizer(self.chain, params[:, 1], params[:, 2])
            psoTrace = np.array([pso.gbest.position.copy() for _ in pso.sample()])
            params[:, 0] = pso.gbest.position
        """

        sampler = CosmoHammerSampler(
                params=params,
                likelihoodComputationChain=self.chain,
                filePrefix='%s' % self.name + '%d' % num,
                walkersRatio=walker_ratio,
                burninIterations=burnin,
                sampleIterations=samples, threadCount=threads)

        print("started sampling:")
        start = time.time()
        sampler.startSampling()
        end = time.time()
        tics = end-start
        print("The time taken %.2f sec. done!" % tics)
        print('Done!')



import numpy as np
import matplotlib.pyplot as plt
from tensorflow import keras as ks

model = ks.models.Sequential()
Dense = ks.layers.Dense


class ANN:

    def __init__(self, data, parameters, name='powerspectrum',
                 epochs=100, batch=10, optimizer='adam',
                 kernel_init='uniform', validation=0.038):
        """
                : data: must be an array (N*k_bins)
                : params: must be an array (N*n_parameters)
                : epochs: number of epochs for the training
                : batch: batch size
                : optimizer: choose the optimizer ('adam','adamax', 'graident_descent'),
                default: 'adam'
                :kernel_init: kernel_initilizier: default is 'uniform'
                :validation: The ration of validation to the training set, default:0.010,
                which is 10 out of 1000 training data, (10 for validation, 9990 for training)

                :return: A trained model

                e.g.: data =array([10,20,40,80,130],....N)
                params = array([10,20,50],...N)
        """
        self.name = name
        self.data = data
        self.parameters = parameters
        self.epochs = epochs
        self.batch = batch
        self.kernel_init = kernel_init
        self.optimizer = optimizer
        self.validation = validation

    def train_pk(self):
        """
        :return: For training ANN based 21-cm powerspectrum
        """

        model.add(Dense(3, input_shape=np.shape(self.parameters[0]),
                        activation='elu', kernel_initializer=self.kernel_init,))
        model.add(Dense(48, activation='elu', kernel_initializer=self.kernel_init))
        model.add(Dense(28, activation='elu', kernel_initializer=self.kernel_init))
        model.add(Dense(14, activation='elu', kernel_initializer=self.kernel_init))
        model.add(Dense(len(self.data[0]), activation='linear'))
        model.compile(loss='mse', optimizer=self.optimizer, metrics=['acc'])
        history = model.fit(self.parameters, self.data,
                            validation_split=self.validation,
                            epochs=self.epochs,
                            batch_size=self.batch)

        model.save('pk.h5')
        np.save('pk_history', history.history)
        print('The model pk.h5 saved!')

    def train_bk(self):
        """
        :return: For training ANN based 21cm Bispectrum
        """
        model.add(Dense(656, input_shape=np.shape(self.parameters[0]), activation='elu'))
        model.add(Dense(328, activation='elu'))
        model.add(Dense(len(self.data[0])))
        optimizer = ks.optimizers.Adam(lr=0.0001, )
        model.compile(loss='mse', optimizer=optimizer, metrics=['acc'], )
        history = model.fit(self.parameters, self.data,
                            validation_split=self.validation,
                            epochs=self.epochs,
                            batch_size=self.batch, )

        model.save('bk.h5')
        np.save('bk_history', history.history)
        print('bk.h5 model saved!')

    def train_bk_model_01(self):    # old version
        """
        :return: for training ANN based 21-cm Bispectrum
        """
        model.add(Dense(80, input_shape=np.shape(self.parameters[0]), activation='elu'))
        model.add(Dense(320, activation='elu'))
        model.add(Dense(460, activation='elu'))
        model.add(Dense(560, activation='elu'))
        model.add(Dense(260, activation='elu'))
        model.add(Dense(100, activation='elu'))
        model.add(Dense(len(self.data[0])))

        optimizer = ks.optimizers.Adamax(lr=0.0001, beta_1=0.2, beta_2=0.088)
        model.compile(loss='mse', optimizer=optimizer, metrics=['acc'], )
        history = model.fit(self.parameters,
                            self.data,
                            validation_split=self.validation,
                            epochs=self.epochs,
                            batch_size=self.batch, )

        model.save('Bk_model_01.h5')
        np.save('Bk_model_01_history', history.history)
        print('Bk_model_01.h5 model saved at current directory')

    def train_bk_model_02(self):    # old version
        """
        Train your Bispectrum data given the parameters
        Please normalize the data before training
        :return: A trained ANN based 21-cm Bispectrum EmuPBk
        """
        model.add(Dense(80, input_shape=np.shape(self.parameters[0]), activation='elu'))
        model.add(Dense(320, activation='elu'))
        model.add(Dense(460, activation='relu'))
        model.add(Dense(560, activation='elu'))
        model.add(Dense(260, activation='elu'))
        model.add(Dense(100, activation='elu'))
        model.add(Dense(len(self.data[0])))
        optimizer = ks.optimizers.Adam(lr=0.0001)
        model.compile(loss='mse',
                      optimizer=optimizer,
                      metrics=['acc'])
        history = model.fit(self.parameters,
                            self.data,
                            validation_split=self.validation,
                            epochs=self.epochs,
                            batch_size=self.batch, )

        model.save('bk_model_02.h5')
        np.save('bk_model_02_history', history.history)
        print('bk_model_02.h5 model saved at current location')

    def get_plot(self, history):
        """
        Please normalize the data before training
        e.g.: for Bispectrum at k1 = 0.2, divide data/100. is a good choice,

        return: gives the plot of model accuracy and model loss for the training set and the test set
        """

        plt.figure(figsize=(16, 6))

        plt.suptitle(r'Accuracy and Loss on the validation set for %s' % self.name)
        plt.subplot(1, 2, 1)
        c = list(np.arange(0, 1, 0.2))
        c.append(0.96)
        plt.yticks(c)
        plt.plot(history.history['acc'], lw=3, color='purple')
        plt.plot(history.history['val_acc'], lw=1, color='yellow', alpha=0.7)
        plt.hlines(y=0.96, xmin=0, xmax=self.epochs, linestyles='dashdot', lw=2)

        plt.title('model accuracy')
        plt.ylabel('accuracy')
        plt.xlabel('epoch')
        plt.legend(['training', 'validation'], loc='lower right')

        plt.subplot(1, 2, 2)

        plt.plot(history.history['loss'], color='purple')
        plt.plot(history.history['val_loss'], color='orange')
        plt.hlines(y=0, xmin=0, xmax=self.epochs, linestyles='dashed', lw=1)
        plt.title('model loss')
        plt.ylabel('loss')
        plt.xlabel('epoch')
        plt.legend(['training', 'validation'], loc='upper right')
        plt.savefig('%s_acc._vs._loss.png' % self.name)
        print('Successfully saved the figure at current location')


import os
import numpy as np
import matplotlib.pyplot as plt
from tensorflow import keras as ks
from chainconsumer import ChainConsumer
from celluloid import Camera
import matplotlib.ticker as ticker
import imageio

path = os.path.abspath(os.path.join(__file__, os.pardir))
path = path+'/MCMC/existing_models/'
ex_models = ['pk.h5', 'bk.h5']
dr = './'
images = []


class AnimatePk:
    """
    Generelized function for animating comparison between
    simulated data(powerspectrum) vs ANN predictions
    """

    def __init__(self, test_data, test_params, k, xh_test, load_model=path+ex_models[0], norm=1.0):
        """
         :param: test_data : test_data(powerspectum)
         :param: test_params : test_parameters of EoR
         :param: k: corresponding k-values
         :param: load_model: load model from a directory path, default: uses existing models.
         :param: norm: normalizing factor, default is no rescaling = 1
         :returns: animation between test_data and ANN predictions

        """

        self.test_data = test_data
        self.test_params = test_params
        self.k = k
        self.xh_test = xh_test
        model = ks.models.load_model(load_model)
        pk_pred = model.predict(test_params)
        self.pk_pred = pk_pred * norm

    def get_animation_pk(self):
        fig = plt.figure(figsize=(10, 7))
        plt.style.use('seaborn')
        plt.grid(color='white', lw=1.5)
        camera = Camera(fig)
        for i in range(15):
            plt.suptitle('EoR 21-cm Powerspectrum, Simulation vs. ANN Prediction',
                         size=20, color='Darkblue')
            plt.xlabel(r'$\mathrm{k(Mpc^{-1})}$', size=20, color='darkblue')
            plt.ylabel(r'$\rm k^3P(k)/2\pi^2(mK^2)$', size=25, color='Darkblue')
            plt.xscale('log')
            plt.ylim(0, int(np.max(self.test_data)+80))
            plt.yscale('symlog')
            plt.plot(self.k, self.pk_pred[i], marker='o', ls='-', color='blue', lw=3)
            plt.plot(self.k, self.test_data[i], ls='--', lw=2, color='red')
            plt.legend(['ANN Prediction',
                        r' Powerspectrum for $\rm \zeta = {0:.2f}, '
                        r'Rmfp = {1:.2f}, Mhmin(10^8 M\odot) = {2:.1f}$'
                       .format(self.test_params[i, 1], self.test_params[i, 2],
                               self.test_params[i, 0], self.xh_test[i])],
                       loc='lower right', prop={'size': 13})
            camera.snap()
        animation = camera.animate()
        animation.save('pk.gif', fps=1, dpi=160)
        print('pk.gif saved!')


class AnimateBk:

    def __init__(self, bk_test, params_test, ntri, xh, load_model=path+ex_models[1],
                 cos_min=0.50, cos_max=0.99, cos_step=0.1,
                 k2byk1_min=0.50, k2byk1_max=1.00, k2byk1_step=0.05, norm=100.):
        """
        simulation specific function, will not be so useful on more than 3 EoR parameters.
        It will give the animation of real bispectrum vs ANN predictions
         for only Unique triangle space configuration.
         :param: test_data : test_data(Bispectrum)
         :param: test_params : test_parameters of EoR
         :param: ntri: number of triangles contributes in each bisepctrum.
         :param: xh : neutral fraction
         :param: load_model: load_your own model from a directory path, (default: uses existing model)
         :param: costheta: provide cos_min and cos_max and its step
         :param: k2byk1: provide the range of k2byk1 and its step
         :param: norm: normalizing factor, (default: 100.)
        """
        costheta = np.array([0.52, 0.57, 0.62, 0.67, 0.72, 0.77, 0.82, 0.87, 0.92, 0.97])
        self.costheta = costheta
        non_na_index = np.array([18, 19, 27, 28, 29, 35, 36, 37, 38, 39, 44, 45, 46,
                                 47, 48, 49, 53, 54, 55, 56, 57, 58, 59, 63, 64, 65,
                                 66, 67, 68, 69, 72, 73, 74, 75, 76, 77, 78, 79, 81,
                                 82, 83, 84, 85, 86, 87, 88, 89, 91, 92, 93, 94, 95,
                                 96, 97, 98, 99, 100, 101, 102, 103, 104, 105, 106, 107, 108, 109])
        self.k1 = np.array([0.1903934, 0.3220935, 0.5448941, 0.9218117, 1.559453])
        self.params_test = params_test
        self.xh = xh
        k2byk1 = np.arange(k2byk1_min, k2byk1_max + k2byk1_step, k2byk1_step)
        costheta = np.arange(cos_min, cos_max + cos_step, cos_step)
        self.costheta = costheta
        self.k2byk1 = np.reshape(k2byk1, (len(k2byk1), 1))
        # cond = costheta.reshape(len(costheta), 1) * k2byk1.reshape(1, len(k2byk1))
        # bk_test_dummy = np.reshape(bk_test, shape = (len(self.params_test), len(k2byk1) * len(costheta)))
        # ntri_dummy = np.reshape(ntri, shape = (len(k2byk1 * len(costheta))))
        model = ks.models.load_model(load_model)
        # costheta = np.arange(cos_min, cos_max + cos_step, cos_step)
        # self.costheta = costheta
        self.k2byk1 = np.reshape(k2byk1, (len(k2byk1), 1))
        cond = costheta.reshape(len(costheta), 1) * k2byk1.reshape(1, len(k2byk1))
        ntri_dummy = np.reshape(ntri, (len(self.k1), len(non_na_index)))
        bk_test_dummy = np.reshape(bk_test, (len(self.params_test), len(self.k1), len(non_na_index)))
        bk_pred_dummy = model.predict(self.params_test)
        bk_pred_dummy = bk_pred_dummy * norm
        bk_pred_dummy = np.reshape(bk_pred_dummy, np.shape(bk_test_dummy))
        bk_test = np.zeros(shape=(len(self.params_test), len(k2byk1) * len(costheta)))
        bk_pdct = np.zeros(shape=bk_test.shape)
        ntri = np.zeros(shape=len(k2byk1) * len(costheta))
        ntri[non_na_index] = ntri_dummy
        for i in range(len(self.xh)):
            bk_pdct[i][non_na_index] = bk_pred_dummy[i]
            bk_test[i][non_na_index] = bk_test_dummy[i]
        bk_test = np.zeros(shape=(len(self.params_test),
                                  len(self.k1), len(k2byk1) * len(costheta)))
        bk_pdct = np.zeros(shape=bk_test.shape)
        ntri = np.zeros(shape=(len(self.k1), len(costheta)*len(k2byk1)))
        for i in range(len(self.k1)):
            ntri[i][non_na_index] = ntri_dummy[i]

        for i in range(len(self.xh)):
            for j in range(len(self.k1)):
                bk_pdct[i][j][non_na_index] = bk_pred_dummy[i][j]
                bk_test[i][j][non_na_index] = bk_test_dummy[i][j]
        bk_test = np.reshape(bk_test, (len(xh), len(self.k1), len(k2byk1), len(costheta)))
        bk_pdct = np.reshape(bk_pdct, np.shape(bk_test))
        ntri = np.reshape(ntri, (len(self.k1), len(k2byk1), len(costheta)))

        for i in range(len(xh)):
            for j in range(len(self.k1)):
                for k in range(len(k2byk1)):
                    for l in range(len(costheta)):
                        if cond.T[k][l] < 0.50:
                            bk_test[i][j][k][l] = np.nan
                            bk_pdct[i][j][k][l] = np.nan
                            ntri[j][k][l] = np.nan

        self.bk_pdct = np.ma.masked_invalid(bk_pdct)
        self.bk_test = np.ma.masked_invalid(bk_test)
        self.ntri = np.ma.masked_invalid(ntri)

    def get_animation_bk(self):
        """
        :returns: Animation between simulated Bispectrum and ANN predictions.
        """
        for i in range(len(self.xh)):
            fig, axes = plt.subplots(nrows=2, ncols=5, figsize=(26, 8), sharex='True', sharey='True')
            b = np.array([self.bk_test[i][0], self.bk_test[i][1], self.bk_test[i][2],
                          self.bk_test[i][3], self.bk_test[i][4],
                          self.bk_pdct[i][0], self.bk_pdct[i][1], self.bk_pdct[i][2],
                          self.bk_pdct[i][3], self.bk_pdct[i][4]])
            fig.text(0.435, 0.96,
                     r' Bispectrum, $\rm \zeta$ = {0:.2f}, Rmfp(Mpc) = {1:.2f},'
                     r' $\rm Mhmin(10^8 M_\odot)$ = {2:.1f}, $\rm xH1$ ={3:.3f}'
                     .format(np.float(self.params_test[i, 1]), np.float(self.params_test[i, 2]),
                             np.float(self.params_test[i, 0]), self.xh[i]),
                     ha='center', size=25, fontstyle='normal')
            for j, ax in zip(b, axes.flat):
                bounds = np.array([-5e3, -2e3, -1e3, 0, 1e3, 2e3, 3e3, 4e3, 5e3, 6e3, 7e3, 8e3, 9e3])
                im = ax.contourf(j, origin='lower', extent=[0.52, .97, 0.50, 1.00],
                                 cmap='seismic', extend='both',
                                 corner_mask=False, levels=bounds)
                fig.text(0.134, 0.9,
                         r'$\mathrm{k_1(Mpc^{-1})}$ = 0.19          '
                         r'$\mathrm{k_1(Mpc^{-1})}$ = 0.32          '
                         r'$\mathrm{k_1(Mpc^{-1})}$ = 0.54           '
                         r'$\mathrm{k_1(Mpc^{-1})}$ = 0.92           '
                         r'$\mathrm{k_1(Mpc^{-1})}$ = 1.55',
                         size=20)
                fig.text(0.435, 0.04, r'$\mathrm{cos(\theta)}$', ha='center', size=30)
                fig.text(0.05, 0.5, r'$\mathrm{k_2/k_1}$', va='center', rotation='vertical', size=30)
                fig.text(0.09, 0.71, r'Simulation', va='center', rotation='vertical', size=25)
                fig.text(0.09, 0.31, r'Emulation', va='center', rotation='vertical', size=25)
                fmt = ticker.ScalarFormatter(useMathText=True)
                fmt.set_powerlimits((0, 0))
                cbar = fig.colorbar(im, label=r"$(k_1^3.k_2^3.B(k_1,k_2,k_3)/(2\pi^2)^2)[mK^3]$",
                                    ax=axes.ravel().tolist(), format=fmt)
                cbar.ax.yaxis.label.set_size(25.)
                plt.savefig('Bk_%d.jpg' % i, dpi=120, bbox_inches='tight')
            fmt = ticker.ScalarFormatter(useMathText=True)
            fmt.set_powerlimits((0, 0))
            cbar = fig.colorbar(im, label=r"$\rm (k_1^3.k_2^3.B(k_1,k_2,k_3)/(2\pi^2)^2)[mK^3]$",
                                ax=axes.ravel().tolist(), format=fmt)
            cbar.ax.yaxis.label.set_size(25.)
            fig.text(0.435, 0.04, r'$\mathrm{cos(\theta)}$', ha='center', size=30)
            fig.text(0.05, 0.5, r'$\mathrm{k_2/k_1}$', va='center', rotation='vertical', size=30)
            fig.text(0.09, 0.71, r'Simulation', va='center', rotation='vertical', size=25)
            fig.text(0.09, 0.31, r'Emulation', va='center', rotation='vertical', size=25)
            plt.savefig('Bk_%d.jpg' % i, dpi=120, bbox_inches='tight')
        for file_name in os.listdir(dr):
            if file_name.endswith('.jpg'):
                file_path = os.path.join(dr, file_name)
                images.append(imageio.imread(file_path))
        imageio.mimsave('Bk.gif', images, fps=1)
        os.system('rm Bk_*.jpg')
    print('Bk.gif saved!')

    def get_bk_vs_cos(self,):
        """
        :return: gives a broad idea of ANN predictions over real
        bispectrum values, at different k2/k1 ratios.
        """
        for j in range(len(self.xh)):
            plt.figure(figsize=(16, 12))
            plt.suptitle(r'$\rm Bispectrum vs. ANN predictions at different k_2/k_1 ratios$',
                         fontsize=15)

            for i in range(1, len(self.k2byk1), 1):
                plt.title(r'$\rm k_2/k_1$ = %.2f' % self.k2byk1[i - 1])
                plt.subplot(3, 4, i)
                plt.xticks([False])
                plt.yticks([False])
                bb = self.bk_test[j][i]
                bp = self.bk_pdct[j][i]
                plt.xlabel(r'cos($\theta$)')
                plt.ylabel(r'$B(k_1,k_2,k_3)$')

                plt.plot(self.costheta, bb, '.-', label='real')
                plt.plot(self.costheta, bp, '.-', label='pred')
                bd = np.subtract(bb, bp)
                plt.plot(self.costheta, bd, '.-', label='diff', )
                plt.legend(loc='upper left')
                plt.title('k2/k1 = %.2f' % self.k2byk1[i])
                filename = 'Bk_vs_cos' + str(j) + '.jpg'
                plt.savefig(filename)
            for file_name in os.listdir(dr):
                if file_name.endswith('.jpg'):
                    file_path = os.path.join(dr, file_name)
                    images.append(imageio.imread(file_path))
        imageio.mimsave('Bk_vs_cos.gif', images, fps=1)
        os.system('rm Bk_vs_cos*.jpg')
        print('Bk_vs_cos.gif animation saved.')


class GetPosterior:
    """
    Returns the posterior distribution
    :params : the posterior parameter array
    walks in the paramter space: type array( N * (no. of parameters) )
    :logl: array contains the values of
    loglikelihood : type array(N*1)
    :name: Powerpsectrum, Bispectrum
    """

    def __init__(self, data):
        self.c = ChainConsumer()
        parameters = [r"$\zeta$", "$R_{mfp}$", r"$Mh_{min} \times 10^8 M\odot$"]
        self.parameters = parameters
        self.data = np.array([data[:, 1], data[:, 2], data[:, 0]])

    def corner(self, index, name='bk', color='blue', dpi=120):
        """
        :param index: index for the plot, type int
        :param name: statistics used, e.g. Powerspectrum, Bispectrum
        :param color: color
        :param dpi: dpi of the plot
        """
        self.c.add_chain(self.data, parameters=self.parameters, color=color)
        self.c.add_chain(self.data, parameters=['zeta', 'rmfp', 'noh'], color=color)

        fig = self.c.plotter.plot(figsize=1.0)

        fig.set_size_inches(3 + fig.get_size_inches())

        fig.savefig('{0:s}_{1:d}.jpg'.format(name, index), dpi=dpi)
        print('{0:s}_{1:d}.jpg saved'.format(name, index))


import os
from tensorflow import keras as ks

path = os.path.abspath(os.path.join(__file__, os.pardir))
path = path+'/MCMC/existing_models/'
ex_models = ['pk.h5', 'bk.h5']


class Predict:
    """
        Epoch of Reionization(EoR) 21-cm powerspectrum and Bispectrum,
        provided astrophysical parameters (Mhmin,Nion,Rmfp)
    """
    def __init__(self, params):
        self.model = ks.models.load_model
        self.params = params

    def existing_models(self, stats):
        """
        stats: int(0, 1)
        choose 0 for powerspectrum, and 1 for bispectrum
        """
        load_model = path+'%s' % ex_models[stats]
        model = self.model(load_model)
        model = model.predict(self.params)
        return model

    def model(self, load_model):
        """
        The 21-cm powerspectrum or 21-cm Bispectrum from k1 range (0.2-1.5) mpc^-1
        based on Artificial Neural Networks(ANN),
            Input:
            :param: params: array([Mhmin,Nion,Rmfp])
            :returns: emulated P(k) or B(k1,k2,k3)
        """

        model = self.model(load_model)
        model = model.predict(self.params)
        return model


import numpy as np


class LikeModule(object):

	"""Likelihood module (CosmoHammer)"""

	def __init__(self, data, nbins, noise, div):
		"""
		:param data: load the observational (test powerspectrum, bispectrum)data, or (test_data, dtype=array,list)
		:param nbins: numbers of k-modes in powerspectrum OR number of triangles for bispectrum(dtype = array, list)
		:param noise: instrumental noise, (dtype = array, list)
		:div: normalizing constant for likelihood.
		"""
		self.data = data
		eye = np.eye(len(data))
		if nbins != 0.:
			cov = abs(data)/np.sqrt(nbins)
			cov =  cov + np.abs(noise)
		else:
			cov = np.abs(noise)
		cov = eye*cov
		self.cov_inv = np.linalg.inv(cov)
		self.div = div

	def computelikelihood(self, ctx):
		"""
		:param ctx: context (load proposed steps from context)
		:return: loglikelihood
		"""
		model_th = ctx.get("model_th")
		diff = np.subtract(model_th, self.data)
		diff = diff.reshape(1, len(self.data))
		logl = -np.dot(diff, np.dot(self.cov_inv, diff.T))/2.
		logl = logl/self.div
		return logl

	@staticmethod
	def setup():

		print("Likelihood setup done!")



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()


setuptools.setup(
    name="EmuPBk",
    version='6.5',
    author='Himanshu Tiwari',
    author_email="himanshuhimang@gmail.com",
    packages=setuptools.find_packages(),
    package_data={'': ['']},
    install_requires=[  "tensorflow", "chainconsumer", "cosmohammer"],
    url="http://github.com/EmuPBk",
    license="MIT License",
    description="ANN based 21-cm Powespectrum and Bispectrum Emulator",
    include_package_data=True,
    keywords=["EmuPBk",
              "ANN emulation on 21-cm powerspectrum and Bispectrum EmuPBk",
              "parameter estimation",
              "cosmology",
              "MCMC"],
    project_urls={
        "Documentation": "https://emupbk.readthedocs.io/en/latest/"},
    classifiers=[
        "Intended Audience :: Science/Research",
        "Programming Language :: Python",
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
)
